#include "spadfs.h"

/* Read various filesystem structures, and check for magic and checksum */
/* TODO:
	check checksum only when reading from disk.
	      use some flag that is reset when reading buffer from disk (which?)
*/

struct txblock *SPADFS_READ_TX_BLOCK(SPADFS *fs, struct buffer_head **bhp, const char *msg)
{
	struct txblock *tb = SPADFS_READ_SECTOR(fs, fs->txb_sec, bhp, msg);
	if (__unlikely(IS_ERR(tb))) return tb;
	if (__unlikely(tb->magic != TXBLOCK_MAGIC)) {
		SPADFS_ERROR(fs, TXFLAGS_FS_ERROR, "BAD MAGIC ON TX BLOCK %LX AT %s", (unsigned long long)fs->txb_sec, msg);
		rel_err:
		spadfs_brelse(fs, *bhp);
		return ERR_PTR(-EFSERROR);
	}
	if (__unlikely(check_checksums(fs))) {
		if (__unlikely(__byte_sum(tb, 512) != CHECKSUM_BASE)) {
			SPADFS_ERROR(fs, TXFLAGS_CHECKSUM_ERROR, "BAD CHECKSUM ON TX BLOCK %LX AT %s", (unsigned long long)fs->txb_sec, msg);
			goto rel_err;
		}
	}
	return tb;
}

/*
 * struct_type means what can we accept:
 *	SRFB_FNODE : fnode_block
 *	SRFB_DNODE : dnode
 *	SRFB_FIXED_FNODE : fixed fnode block
 * if more are set, the type of returned structure must be determined by
 * the caller from magic.
 */

struct fnode_block *SPADFS_READ_FNODE_BLOCK(SPADFS *fs, __d_off secno, struct buffer_head **bhp, int struct_type, const char *msg)
{
	struct fnode_block *fnode_block = SPADFS_READ_SECTOR(fs, secno, bhp, msg);
	if (__unlikely(IS_ERR(fnode_block))) return fnode_block;
	if (__likely(fnode_block->magic == FNODE_BLOCK_MAGIC)) {
		if (__unlikely(!(struct_type & SRFB_FNODE))) goto bad_magic;
		check_fn:
		if (__unlikely(check_checksums(fs)) && __likely(fnode_block->flags & FNODE_BLOCK_CHECKSUM_VALID)) {
			start_concurrent_atomic_buffer_read(*bhp);
			if (__likely(fnode_block->flags & FNODE_BLOCK_CHECKSUM_VALID)) {
				if (__unlikely(__byte_sum(fnode_block, 512) != CHECKSUM_BASE)) {
					end_concurrent_atomic_buffer_read(*bhp);
					SPADFS_ERROR(fs, TXFLAGS_CHECKSUM_ERROR, "BAD CHECKSUM ON FNODE BLOCK %LX AT %s", (unsigned long long)secno, msg);
					goto rel_err;
				}
			}
			end_concurrent_atomic_buffer_read(*bhp);
		}
		return fnode_block;
	}
	if (__likely(fnode_block->magic == DNODE_PAGE_MAGIC)) {
		if (__unlikely(!(struct_type & SRFB_DNODE))) goto bad_magic;
		if (__unlikely((((struct dnode_page *)fnode_block)->gflags & DNODE_GFLAGS_PAGE_SIZE_BITS_MINUS_1) >> DNODE_GFLAGS_PAGE_SIZE_BITS_MINUS_1_SHIFT != fs->sectors_per_page_bits - 1)) {
			SPADFS_ERROR(fs, TXFLAGS_FS_ERROR, "DNODE %LX HAS INVALID PAGE SIZE AT %s", (unsigned long long)secno, msg);
			goto rel_err;
		}
		return fnode_block;
	}
	if (__likely(fnode_block->magic == FIXED_FNODE_BLOCK_MAGIC)) {
		if (__unlikely(!(struct_type & SRFB_FIXED_FNODE))) goto bad_magic;
		goto check_fn;
	}
	bad_magic:
	SPADFS_ERROR(fs, TXFLAGS_FS_ERROR, "BAD MAGIC %08X ON FNODE BLOCK %LX AT %s", fnode_block->magic, (unsigned long long)secno, msg);
	rel_err:
	spadfs_brelse(fs, *bhp);
	return ERR_PTR(-EFSERROR);
}

struct anode *SPADFS_READ_ANODE(SPADFS *fs, __d_off secno, struct buffer_head **bhp, unsigned *vx, const char *msg)
{
	struct anode *anode = SPADFS_READ_SECTOR(fs, secno, bhp, msg);
	if (__unlikely(IS_ERR(anode))) return anode;
	if (__unlikely(anode->magic != ANODE_MAGIC)) {
		SPADFS_ERROR(fs, TXFLAGS_FS_ERROR, "BAD MAGIC ON ANODE %LX AT %s", (unsigned long long)secno, msg);
		rel_err:
		spadfs_brelse(fs, *bhp);
		return ERR_PTR(-EFSERROR);
	}
	if (__unlikely(check_checksums(fs)) && __likely(anode->flags & ANODE_CHECKSUM_VALID)) {
		if (__unlikely(__byte_sum(anode, 512) != CHECKSUM_BASE)) {
			SPADFS_ERROR(fs, TXFLAGS_CHECKSUM_ERROR, "BAD CHECKSUM ON ANODE %LX AT %s", (unsigned long long)secno, msg);
			goto rel_err;
		}
	}
	*vx = anode->valid_extents;
	if (__unlikely((unsigned)(*vx - 1) >= ANODE_N_EXTENTS)) {
		SPADFS_ERROR(fs, TXFLAGS_FS_ERROR, "ANODE %LX HAS %u VALID EXTENTS AT %s", (unsigned long long)secno, (unsigned)anode->valid_extents, msg);
		goto rel_err;
	}
	return anode;
}

